from .libindy import do_call, create_cb

from ctypes import *
from typing import Optional

import logging


async def create_wallet(config: str,
                        credentials: str) -> None:
    """
    Creates a new secure wallet with the given unique name.

    :param config: Wallet configuration json.
     {
       "id": string, Identifier of the wallet.
             Configured storage uses this identifier to lookup exact wallet data placement.
       "storage_type": optional<string>, Type of the wallet storage. Defaults to 'default'.
                      'Default' storage type allows to store wallet data in the local file.
                      Custom storage types can be registered with indy_register_wallet_storage call.
       "storage_config": optional<object>, Storage configuration json. Storage type defines set of supported keys.
                         Can be optional if storage supports default configuration.
                          For 'default' storage type configuration is:
       {
         "path": optional<string>, Path to the directory with wallet files.
                 Defaults to $HOME/.indy_client/wallet.
                 Wallet will be stored in the file {path}/{id}/sqlite.db
       }
     }
    :param credentials: Wallet credentials json
     {
       "key": string, Key or passphrase used for wallet key derivation.
                      Look to key_derivation_method param for information about supported key derivation methods.
       "storage_credentials": optional<object> Credentials for wallet storage. Storage type defines set of supported keys.
                              Can be optional if storage supports default configuration.
                               For 'default' storage type should be empty.
       "key_derivation_method": optional<string> Algorithm to use for wallet key derivation:
                                ARGON2I_MOD - derive secured wallet master key (used by default)
                                ARGON2I_INT - derive secured wallet master key (less secured but faster)
                                RAW - raw wallet key master provided (skip derivation).
                                      RAW keys can be generated with generate_wallet_key call
     }
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_wallet: >>> config: %r, credentials: %r",
                 config,
                 credentials)

    if not hasattr(create_wallet, "cb"):
        logger.debug("create_wallet: Creating callback")
        create_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_config = c_char_p(config.encode('utf-8'))
    c_credentials = c_char_p(credentials.encode('utf-8'))

    await do_call('indy_create_wallet',
                  c_config,
                  c_credentials,
                  create_wallet.cb)

    logger.debug("create_wallet: <<<")


async def open_wallet(config: str,
                      credentials: str) -> int:
    """
    Opens the wallet with specific name.
    Wallet with corresponded name must be previously created with indy_create_wallet method.
    It is impossible to open wallet with the same name more than once.

    :param name: Name of the wallet.
    :param config: Wallet configuration json.
    {
       "id": string, Identifier of the wallet.
             Configured storage uses this identifier to lookup exact wallet data placement.
       "storage_type": optional<string>, Type of the wallet storage. Defaults to 'default'.
                       'Default' storage type allows to store wallet data in the local file.
                       Custom storage types can be registered with indy_register_wallet_storage call.
       "storage_config": optional<object>, Storage configuration json. Storage type defines set of supported keys.
                         Can be optional if storage supports default configuration.
                          For 'default' storage type configuration is:
           {
              "path": optional<string>, Path to the directory with wallet files.
                      Defaults to $HOME/.indy_client/wallet.
                      Wallet will be stored in the file {path}/{id}/sqlite.db
           }

    }
    :param credentials: Wallet credentials json
    {
       "key": string, Key or passphrase used for wallet key derivation.
                      Look to key_derivation_method param for information about supported key derivation methods.
       "rekey": optional<string>, If present, then wallet master key will be rotated to a new one.
       "storage_credentials": optional<object> Credentials for wallet storage. Storage type defines set of supported keys.
                              Can be optional if storage supports default configuration.
                              For 'default' storage type should be empty.
       "key_derivation_method": optional<string> Algorithm to use for wallet key derivation:
                               ARGON2I_MOD - derive secured wallet master key (used by default)
                               ARGON2I_INT - derive secured wallet master key (less secured but faster)
                               RAW - raw wallet master key provided (skip derivation)
       "rekey_derivation_method": optional<string> algorithm to use for master rekey derivation:
                               ARGON2I_MOD - derive secured wallet master rekey (used by default)
                               ARGON2I_INT - derive secured wallet master rekey (less secured but faster)
                               RAW - raw wallet rekey master provided (skip derivation).
                                     RAW keys can be generated with generate_wallet_key call
    }
    :return: Handle to opened wallet to use in methods that require wallet access.
    """

    logger = logging.getLogger(__name__)
    logger.debug("open_wallet: >>> config: %r, credentials: %r",
                 config,
                 credentials)

    if not hasattr(open_wallet, "cb"):
        logger.debug("open_wallet: Creating callback")
        open_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_int32))

    c_config = c_char_p(config.encode('utf-8'))
    c_credentials = c_char_p(credentials.encode('utf-8'))

    res = await do_call('indy_open_wallet',
                        c_config,
                        c_credentials,
                        open_wallet.cb)

    logger.debug("open_wallet: <<< res: %r", res)
    return res


async def close_wallet(handle: int) -> None:
    """
    Closes opened wallet and frees allocated resources.

    :param handle: wallet handle returned by indy_open_wallet.
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("close_wallet: >>> handle: %i", handle)

    if not hasattr(close_wallet, "cb"):
        logger.debug("close_wallet: Creating callback")
        close_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_handle = c_int32(handle)

    await do_call('indy_close_wallet',
                  c_handle,
                  close_wallet.cb)

    logger.debug("close_wallet: <<<")


async def delete_wallet(config: str,
                        credentials: str) -> None:
    """
    Deletes created wallet.

    :param config: Wallet configuration json.
     {
       "id": string, Identifier of the wallet.
             Configured storage uses this identifier to lookup exact wallet data placement.
       "storage_type": optional<string>, Type of the wallet storage. Defaults to 'default'.
                      'Default' storage type allows to store wallet data in the local file.
                      Custom storage types can be registered with indy_register_wallet_storage call.
       "storage_config": optional<object>, Storage configuration json. Storage type defines set of supported keys.
                         Can be optional if storage supports default configuration.
                          For 'default' storage type configuration is:
       {
         "path": optional<string>, Path to the directory with wallet files.
                 Defaults to $HOME/.indy_client/wallet.
                 Wallet will be stored in the file {path}/{id}/sqlite.db
       }
     }
    :param credentials: Wallet credentials json
     {
       "key": string, Key or passphrase used for wallet key derivation.
                      Look to key_derivation_method param for information about supported key derivation methods.
       "storage_credentials": optional<object> Credentials for wallet storage. Storage type defines set of supported keys.
                              Can be optional if storage supports default configuration.
                              For 'default' storage type should be empty.
       "key_derivation_method": optional<string> Algorithm to use for wallet key derivation:
                                ARGON2I_MOD - derive secured wallet master key (used by default)
                                ARGON2I_INT - derive secured wallet master key (less secured but faster)
                                RAW - raw wallet key master provided (skip derivation).
                                      RAW keys can be generated with generate_wallet_key call
     }
    :return:
    """

    logger = logging.getLogger(__name__)
    logger.debug("delete_wallet: >>> config: %r, credentials: %r",
                 config,
                 credentials)

    if not hasattr(delete_wallet, "cb"):
        logger.debug("delete_wallet: Creating callback")
        delete_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_config = c_char_p(config.encode('utf-8'))
    c_credentials = c_char_p(credentials.encode('utf-8'))

    await do_call('indy_delete_wallet',
                  c_config,
                  c_credentials,
                  delete_wallet.cb)

    logger.debug("delete_wallet: <<<")


async def export_wallet(handle: int,
                        export_config_json: str) -> None:
    """
    Exports opened wallet to the file.

    :param handle: wallet handle returned by indy_open_wallet.
    :param export_config_json: JSON containing settings for input operation.
       {
          "path": path of the file that contains exported wallet content
          "key": string, Key or passphrase used for wallet export key derivation.
                         Look to key_derivation_method param for information about supported key derivation methods.
          "key_derivation_method": optional<string> algorithm to use for export key derivation:
                                ARGON2I_MOD - derive secured wallet export key (used by default)
                                ARGON2I_INT - derive secured wallet export key (less secured but faster)
                                RAW - raw wallet export key provided (skip derivation).
                                      RAW keys can be generated with generate_wallet_key call
       }
    :return:
    """

    logger = logging.getLogger(__name__)
    logger.debug("export_wallet: >>> handle: %r, export_config_json: %r",
                 handle,
                 export_config_json)

    if not hasattr(export_wallet, "cb"):
        logger.debug("export_wallet: Creating callback")
        export_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_export_config_json = c_char_p(export_config_json.encode('utf-8'))

    await do_call('indy_export_wallet',
                  handle,
                  c_export_config_json,
                  export_wallet.cb)

    logger.debug("export_wallet: <<<")


async def import_wallet(config: str,
                        credentials: str,
                        import_config_json: str) -> None:
    """
    Creates a new secure wallet with the given unique name and then imports its content
    according to fields provided in import_config
    This can be seen as an indy_create_wallet call with additional content import

    :param config: Wallet configuration json.
     {
       "id": string, Identifier of the wallet.
             Configured storage uses this identifier to lookup exact wallet data placement.
       "storage_type": optional<string>, Type of the wallet storage. Defaults to 'default'.
                      'Default' storage type allows to store wallet data in the local file.
                      Custom storage types can be registered with indy_register_wallet_storage call.
       "storage_config": optional<object>, Storage configuration json. Storage type defines set of supported keys.
                         Can be optional if storage supports default configuration.
                         For 'default' storage type configuration is:
       {
         "path": optional<string>, Path to the directory with wallet files.
                 Defaults to $HOME/.indy_client/wallet.
                 Wallet will be stored in the file {path}/{id}/sqlite.db
       }
     }
    :param credentials: Wallet credentials json
     {
       "key": string, Key or passphrase used for wallet key derivation.
                      Look to key_derivation_method param for information about supported key derivation methods.
       "storage_credentials": optional<object> Credentials for wallet storage. Storage type defines set of supported keys.
                              Can be optional if storage supports default configuration.
                              For 'default' storage type should be empty.
       "key_derivation_method": optional<string> Algorithm to use for wallet key derivation:
                                 ARGON2I_MOD - derive secured wallet master key (used by default)
                                 ARGON2I_INT - derive secured wallet master key (less secured but faster)
                                 RAW - raw wallet key master provided (skip derivation).
                                       RAW keys can be generated with generate_wallet_key call
     }
    :param import_config_json: JSON containing settings for input operationЖ {
     "path": path of the file that contains exported wallet content
     "key": key used for export of the wallet
   }
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("import_wallet: >>> config: %r, credentials: %r, import_config_json: %r",
                 config,
                 credentials,
                 import_config_json)

    if not hasattr(import_wallet, "cb"):
        logger.debug("import_wallet: Creating callback")
        import_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_config = c_char_p(config.encode('utf-8'))
    c_credentials = c_char_p(credentials.encode('utf-8'))
    c_import_config_json = c_char_p(import_config_json.encode('utf-8'))

    await do_call('indy_import_wallet',
                  c_config,
                  c_credentials,
                  c_import_config_json,
                  import_wallet.cb)

    logger.debug("import_wallet: <<<")


async def generate_wallet_key(config: Optional[str]) -> str:
    """
    Generate wallet master key.
    Returned key is compatible with "RAW" key derivation method.
    It allows to avoid expensive key derivation for use cases when wallet keys can be stored in a secure enclave.

    :param config: (optional) key configuration json.
     {
        "seed": string, (optional) Seed that allows deterministic key creation (if not set random one will be created).
                                   Can be UTF-8, base64 or hex string.
     }
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("generate_wallet_key: >>> config: %r",
                 config)

    if not hasattr(generate_wallet_key, "cb"):
        logger.debug("generate_wallet_key: Creating callback")
        generate_wallet_key.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_config = c_char_p(config.encode('utf-8')) if config is not None else None

    key = await do_call('indy_generate_wallet_key',
                        c_config,
                        generate_wallet_key.cb)

    res = key.decode()

    logger.debug("generate_wallet_key: <<< res: %r", res)
    return res



from typing import Optional

from .libindy import do_call, create_cb

from ctypes import *

import logging


async def create_and_store_my_did(wallet_handle: int,
                                  did_json: str) -> (str, str):
    """
    Creates keys (signing and encryption keys) for a new
    DID (owned by the caller of the library).
    Identity's DID must be either explicitly provided, or taken as the first 16 bit of verkey.
    Saves the Identity DID with keys in a secured Wallet, so that it can be used to sign
    and encrypt transactions.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did_json: Identity information as json. Example:
        {
            "did": string, (optional;
                    if not provided and cid param is false then the first 16 bit of the verkey will be
                    used as a new DID;
                    if not provided and cid is true then the full verkey will be used as a new DID;
                    if provided, then keys will be replaced - key rotation use case)
	        "seed": string, (optional) Seed that allows deterministic key creation (if not set random one will be created).
	                                    Can be UTF-8, base64 or hex string.
            "crypto_type": string, (optional; if not set then ed25519 curve is used;
                      currently only 'ed25519' value is supported for this field)
            "cid": bool, (optional; if not set then false is used;)
            "method_name": string, (optional) method name to create fully qualified did.
        }
    :return: DID and verkey (for verification of signature)
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_and_store_my_did: >>> wallet_handle: %r, did_json: %r",
                 wallet_handle,
                 did_json)

    if not hasattr(create_and_store_my_did, "cb"):
        logger.debug("create_wallet: Creating callback")
        create_and_store_my_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did_json = c_char_p(did_json.encode('utf-8'))

    did, verkey = await do_call('indy_create_and_store_my_did',
                                c_wallet_handle,
                                c_did_json,
                                create_and_store_my_did.cb)

    res = (did.decode(), verkey.decode())

    logger.debug("create_and_store_my_did: <<< res: %r", res)
    return res


async def replace_keys_start(wallet_handle: int,
                             did: str,
                             identity_json: str) -> str:
    """
    Generated new keys (signing and encryption keys) for an existing
    DID (owned by the caller of the library).

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did: signing DID
    :param identity_json: Identity information as json. Example:
        {
	        "seed": string, (optional) Seed that allows deterministic key creation (if not set random one will be created).
	                                   Can be UTF-8, base64 or hex string.
            "crypto_type": string, (optional; if not set then ed25519 curve is used;
                      currently only 'ed25519' value is supported for this field)
        }
    :return: verkey
    """

    logger = logging.getLogger(__name__)
    logger.debug("replace_keys_start: >>> wallet_handle: %r, did: %r, identity_json: %r",
                 wallet_handle,
                 did,
                 identity_json)

    if not hasattr(replace_keys_start, "cb"):
        logger.debug("replace_keys_start: Creating callback")
        replace_keys_start.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_identity_json = c_char_p(identity_json.encode('utf-8'))

    verkey = await do_call('indy_replace_keys_start',
                           c_wallet_handle,
                           c_did,
                           c_identity_json,
                           replace_keys_start.cb)

    res = verkey.decode()

    logger.debug("replace_keys_start: <<< res: %r", res)
    return res


async def replace_keys_apply(wallet_handle: int,
                             did: str) -> None:
    """
    Apply temporary keys as main for an existing DID (owned by the caller of the library).

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did: The DID to resolve key.
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("replace_keys_apply: >>> wallet_handle: %r, did: %r",
                 wallet_handle,
                 did)

    if not hasattr(replace_keys_apply, "cb"):
        logger.debug("replace_keys_apply: Creating callback")
        replace_keys_apply.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    await do_call('indy_replace_keys_apply',
                  c_wallet_handle,
                  c_did,
                  replace_keys_apply.cb)

    logger.debug("replace_keys_apply: <<<")


async def store_their_did(wallet_handle: int,
                          identity_json: str) -> None:
    """
    Saves their DID for a pairwise connection in a secured Wallet,
    so that it can be used to verify transaction.
    Updates DID associated verkey in case DID already exists in the Wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param identity_json: Identity information as json. Example:
        {
           "did": string, (required)
           "verkey": string (optional, if only pk is provided),
           "crypto_type": string, (optional; if not set then ed25519 curve is used;
                  currently only 'ed25519' value is supported for this field)
        }
    :return: None
    """

    logger = logging.getLogger(__name__)
    logger.debug("store_their_did: >>> wallet_handle: %r, identity_json: %r",
                 wallet_handle,
                 identity_json)

    if not hasattr(store_their_did, "cb"):
        logger.debug("store_their_did: Creating callback")
        store_their_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_identity_json = c_char_p(identity_json.encode('utf-8'))

    res = await do_call('indy_store_their_did',
                        c_wallet_handle,
                        c_identity_json,
                        store_their_did.cb)

    logger.debug("store_their_did: <<< res: %r", res)
    return res


async def create_key(wallet_handle: int,
                     key_json: str) -> str:
    """
    Creates keys pair and stores in the wallet.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param key_json: Key information as json. Example:
        {
	        "seed": string, (optional) Seed that allows deterministic key creation (if not set random one will be created).
	                                   Can be UTF-8, base64 or hex string.
            "crypto_type": string, // Optional (if not set then ed25519 curve is used);
                    Currently only 'ed25519' value is supported for this field.
        }
    :return: verkey: Ver key of generated key pair, also used as key identifier
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_key: >>> wallet_handle: %r, key_json: %r",
                 wallet_handle,
                 key_json)

    if not hasattr(create_key, "cb"):
        logger.debug("create_key: Creating callback")
        create_key.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_key_json = c_char_p(key_json.encode('utf-8'))

    verkey = await do_call('indy_create_key',
                           c_wallet_handle,
                           c_key_json,
                           create_key.cb)

    res = verkey.decode()

    logger.debug("create_key: <<< res: %r", res)
    return res


async def set_key_metadata(wallet_handle: int,
                           verkey: str,
                           metadata: str) -> None:
    """
    Creates keys pair and stores in the wallet.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param verkey: the key (verkey, key id) to store metadata.
    :param metadata: the meta information that will be store with the key.
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("set_key_metadata: >>> wallet_handle: %r, verkey: %r, metadata: %r",
                 wallet_handle,
                 verkey,
                 metadata)

    if not hasattr(set_key_metadata, "cb"):
        logger.debug("set_key_metadata: Creating callback")
        set_key_metadata.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_verkey = c_char_p(verkey.encode('utf-8'))
    c_metadata = c_char_p(metadata.encode('utf-8'))

    await do_call('indy_set_key_metadata',
                  c_wallet_handle,
                  c_verkey,
                  c_metadata,
                  set_key_metadata.cb)

    logger.debug("create_key: <<<")


async def get_key_metadata(wallet_handle: int,
                           verkey: str) -> str:
    """
    Retrieves the meta information for the giving key in the wallet.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param verkey: The key (verkey, key id) to retrieve metadata.
    :return: metadata: The meta information stored with the key; Can be null if no metadata was saved for this key.
    """

    logger = logging.getLogger(__name__)
    logger.debug("get_key_metadata: >>> wallet_handle: %r, verkey: %r",
                 wallet_handle,
                 verkey)

    if not hasattr(get_key_metadata, "cb"):
        logger.debug("get_key_metadata: Creating callback")
        get_key_metadata.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_verkey = c_char_p(verkey.encode('utf-8'))

    metadata = await do_call('indy_get_key_metadata',
                             c_wallet_handle,
                             c_verkey,
                             get_key_metadata.cb)

    res = metadata.decode()

    logger.debug("get_key_metadata: <<< res: %r", res)
    return res


async def key_for_did(pool_handle: int,
                      wallet_handle: int,
                      did: str) -> str:
    """
    Returns ver key (key id) for the given DID.

    "key_for_did" call follow the idea that we resolve information about their DID from
    the ledger with cache in the local wallet. The "open_wallet" call has freshness parameter
    that is used for checking the freshness of cached pool value.

    Note if you don't want to resolve their DID info from the ledger you can use
    "key_for_local_did" call instead that will look only to local wallet and skip
    freshness checking.

    Note that "create_and_store_my_did" makes similar wallet record as "create_key".
    As result we can use returned ver key in all generic crypto and messaging functions.

    :param pool_handle: Pool handle (created by open_pool).
    :param wallet_handle: Wallet handle (created by open_wallet).
    :param did: The DID to resolve key.
    :return: key:   The DIDs ver key (key id).
    """

    logger = logging.getLogger(__name__)
    logger.debug("key_for_did: >>> pool_handle: %r, wallet_handle: %r, did: %r",
                 pool_handle,
                 wallet_handle,
                 did)

    if not hasattr(key_for_did, "cb"):
        logger.debug("key_for_did: Creating callback")
        key_for_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_pool_handle = c_int32(pool_handle)
    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    key = await do_call('indy_key_for_did',
                        c_pool_handle,
                        c_wallet_handle,
                        c_did,
                        key_for_did.cb)

    res = key.decode()

    logger.debug("key_for_did: <<< res: %r", res)
    return res


async def key_for_local_did(wallet_handle: int,
                            did: str) -> str:
    """
    Returns ver key (key id) for the given DID.

    "key_for_local_did" call looks data stored in the local wallet only and skips freshness checking.

    Note if you want to get fresh data from the ledger you can use "key_for_did" call
    instead.

    Note that "create_and_store_my_did" makes similar wallet record as "create_key".
    As result we can use returned ver key in all generic crypto and messaging functions.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param did: The DID to resolve key.
    :return: key: The DIDs ver key (key id).
    """

    logger = logging.getLogger(__name__)
    logger.debug("key_for_local_did: >>> wallet_handle: %r, did: %r",
                 wallet_handle,
                 did)

    if not hasattr(key_for_local_did, "cb"):
        logger.debug("key_for_local_did: Creating callback")
        key_for_local_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    key = await do_call('indy_key_for_local_did',
                        c_wallet_handle,
                        c_did,
                        key_for_local_did.cb)

    res = key.decode()

    logger.debug("key_for_local_did: <<< res: %r", res)
    return res


async def set_endpoint_for_did(wallet_handle: int,
                               did: str,
                               address: str,
                               transport_key: str) -> None:
    """
    Set/replaces endpoint information for the given DID.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param did: The DID to resolve endpoint.
    :param address: The DIDs endpoint address.
    :param transport_key: The DIDs transport key (ver key, key id).
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("set_endpoint_for_did: >>> wallet_handle: %r, did: %r, address: %r, transport_key: %r",
                 wallet_handle,
                 did,
                 address,
                 transport_key)

    if not hasattr(set_endpoint_for_did, "cb"):
        logger.debug("set_endpoint_for_did: Creating callback")
        set_endpoint_for_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_address = c_char_p(address.encode('utf-8'))
    c_transport_key = c_char_p(transport_key.encode('utf-8'))

    await do_call('indy_set_endpoint_for_did',
                  c_wallet_handle,
                  c_did,
                  c_address,
                  c_transport_key,
                  set_endpoint_for_did.cb)

    logger.debug("set_endpoint_for_did: <<<")


async def get_endpoint_for_did(wallet_handle: int,
                               pool_handle: int,
                               did: str) -> (str, Optional[str]):
    """
    Returns endpoint information for the given DID.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param pool_handle: Pool handle (created by open_pool).
    :param did: The DID to resolve endpoint.
    :return: (endpoint, transport_vk)
    """

    logger = logging.getLogger(__name__)
    logger.debug("get_endpoint_for_did: >>> wallet_handle: %r, pool_handle: %r, did: %r",
                 wallet_handle,
                 pool_handle,
                 did)

    if not hasattr(get_endpoint_for_did, "cb"):
        logger.debug("get_endpoint_for_did: Creating callback")
        get_endpoint_for_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_pool_handle = c_int32(pool_handle)
    c_did = c_char_p(did.encode('utf-8'))

    endpoint, transport_vk = await do_call('indy_get_endpoint_for_did',
                                           c_wallet_handle,
                                           c_pool_handle,
                                           c_did,
                                           get_endpoint_for_did.cb)

    endpoint = endpoint.decode()
    transport_vk = transport_vk.decode() if transport_vk is not None else None
    res = (endpoint, transport_vk)

    logger.debug("get_endpoint_for_did: <<< res: %r", res)
    return res


async def set_did_metadata(wallet_handle: int,
                           did: str,
                           metadata: str) -> None:
    """
    Saves/replaces the meta information for the giving DID in the wallet.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param did: the DID to store metadata.
    :param metadata: the meta information that will be store with the DID.
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("set_did_metadata: >>> wallet_handle: %r, did: %r, metadata: %r",
                 wallet_handle,
                 did,
                 metadata)

    if not hasattr(set_did_metadata, "cb"):
        logger.debug("set_did_metadata: Creating callback")
        set_did_metadata.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_metadata = c_char_p(metadata.encode('utf-8'))

    await do_call('indy_set_did_metadata',
                  c_wallet_handle,
                  c_did,
                  c_metadata,
                  set_did_metadata.cb)

    logger.debug("set_did_metadata: <<<")


async def get_did_metadata(wallet_handle: int,
                           did: str) -> str:
    """
    Retrieves the meta information for the giving DID in the wallet.

    :param wallet_handle: Wallet handle (created by open_wallet).
    :param did: The DID to retrieve metadata.
    :return: metadata: The meta information stored with the DID; Can be null if no metadata was saved for this DID.
    """

    logger = logging.getLogger(__name__)
    logger.debug("get_did_metadata: >>> wallet_handle: %r, did: %r",
                 wallet_handle,
                 did)

    if not hasattr(get_did_metadata, "cb"):
        logger.debug("get_did_metadata: Creating callback")
        get_did_metadata.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    metadata = await do_call('indy_get_did_metadata',
                             c_wallet_handle,
                             c_did,
                             get_did_metadata.cb)

    res = metadata.decode()

    logger.debug("get_did_metadata: <<< res: %r", res)
    return res


async def get_my_did_with_meta(wallet_handle: int, did: str) -> str:
    """
    Get DID metadata and verkey stored in the wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did: The DID to retrieve metadata.
    :return: DID with verkey and metadata.
    """

    logger = logging.getLogger(__name__)
    logger.debug("get_my_did_with_meta: >>> wallet_handle: %r, did: %r",
                 wallet_handle,
                 did)

    if not hasattr(get_my_did_with_meta, "cb"):
        logger.debug("get_my_did_with_meta: Creating callback")
        get_my_did_with_meta.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    did_with_meta = await do_call('indy_get_my_did_with_meta',
                                  c_wallet_handle,
                                  c_did,
                                  get_my_did_with_meta.cb)

    res = did_with_meta.decode()

    logger.debug("get_my_did_with_meta: <<< res: %r", res)
    return res


async def list_my_dids_with_meta(wallet_handle: int) -> str:
    """
    List DIDs and metadata stored in the wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :return: List of DIDs with verkeys and meta data.
    """

    logger = logging.getLogger(__name__)
    logger.debug("list_my_dids_with_meta: >>> wallet_handle: %r",
                 wallet_handle)

    if not hasattr(list_my_dids_with_meta, "cb"):
        logger.debug("list_my_dids_with_meta: Creating callback")
        list_my_dids_with_meta.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)

    dids_with_meta = await do_call('indy_list_my_dids_with_meta',
                                   c_wallet_handle,
                                   list_my_dids_with_meta.cb)

    res = dids_with_meta.decode()

    logger.debug("list_my_dids_with_meta: <<< res: %r", res)
    return res


async def abbreviate_verkey(did: str,
                            full_verkey: str) -> str:
    """
    Retrieves abbreviated verkey if it is possible otherwise return full verkey.

    :param did: The DID.
    :param full_verkey: The DIDs verification key,
    :return: metadata: Either abbreviated or full verkey.
    """

    logger = logging.getLogger(__name__)
    logger.debug("abbreviate_verkey: >>> did: %r, full_verkey: %r",
                 did, full_verkey)

    if not hasattr(abbreviate_verkey, "cb"):
        logger.debug("abbreviate_verkey: Creating callback")
        abbreviate_verkey.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_did = c_char_p(did.encode('utf-8'))
    c_full_verkey = c_char_p(full_verkey.encode('utf-8'))

    metadata = await do_call('indy_abbreviate_verkey',
                             c_did,
                             c_full_verkey,
                             abbreviate_verkey.cb)

    res = metadata.decode()

    logger.debug("abbreviate_verkey: <<< res: %r", res)
    return res


async def qualify_did(wallet_handle: int,
                      did: str,
                      method: str) -> str:
    """
    Update DID stored in the wallet to make fully qualified, or to do other DID maintenance.
        - If the DID has no prefix, a prefix will be appended (prepend did:peer to a legacy did)
        - If the DID has a prefix, a prefix will be updated (migrate did:peer to did:peer-new)

    Update DID related entities stored in the wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did: target DID stored in the wallet.
    :param method: method to apply to the DID.
    :return: fully qualified did
    """

    logger = logging.getLogger(__name__)
    logger.debug("qualify_did: >>> wallet_handle: %r, did: %r, method: %r",
                 wallet_handle, did, method)

    if not hasattr(qualify_did, "cb"):
        logger.debug("qualify_did: Creating callback")
        qualify_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_method = c_char_p(method.encode('utf-8'))

    full_qualified_did = await do_call('indy_qualify_did',
                                       c_wallet_handle,
                                       c_did,
                                       c_method,
                                       qualify_did.cb)

    res = full_qualified_did.decode()

    logger.debug("qualify_did: <<< res: %r", res)
    return res



from .libindy import do_call, create_cb

from typing import Optional
from ctypes import *

import logging


async def create_payment_address(wallet_handle: int,
                                 payment_method: str,
                                 config: str) -> str:
    """
     Create the payment address for specified payment method


     This method generates private part of payment address
     and stores it in a secure place. Ideally it should be
     secret in libindy wallet (see crypto module).

     Note that payment method should be able to resolve this
     secret by fully resolvable payment address format.

    :param wallet_handle: wallet handle (created by open_wallet).
    :param payment_method: Payment method to use (for example, 'sov').
    :param config: payment address config as json:
       {
         seed: <str>, // allows deterministic creation of payment address
       }
    :return: payment_address: public identifier of payment address in fully resolvable payment address format.
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_payment_address: >>> wallet_handle: %r, payment_method: %r, config: %r",
                 wallet_handle,
                 payment_method,
                 config)

    if not hasattr(create_payment_address, "cb"):
        logger.debug("create_payment_address: Creating callback")
        create_payment_address.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    config = c_char_p(config.encode('utf-8'))

    request_result = await do_call('indy_create_payment_address',
                                   c_wallet_handle,
                                   c_payment_method,
                                   config,
                                   create_payment_address.cb)

    res = request_result.decode()
    logger.debug("create_payment_address: <<< res: %r", res)
    return res


async def list_payment_addresses(wallet_handle: int) -> str:
    """
     Lists all payment addresses that are stored in the wallet

    :param wallet_handle: wallet handle (created by open_wallet).
    :return: payment_addresses_json: json array of string with json addresses
    """

    logger = logging.getLogger(__name__)
    logger.debug("list_payment_addresses: >>> wallet_handle: %r",
                 wallet_handle)

    if not hasattr(list_payment_addresses, "cb"):
        logger.debug("list_payment_addresses: Creating callback")
        list_payment_addresses.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)

    request_result = await do_call('indy_list_payment_addresses',
                                   c_wallet_handle,
                                   list_payment_addresses.cb)

    res = request_result.decode()
    logger.debug("list_payment_addresses: <<< res: %r", res)
    return res


async def add_request_fees(wallet_handle: int,
                           submitter_did: str,
                           req_json: str,
                           inputs_json: str,
                           outputs_json: str,
                           extra: Optional[str]) -> (str, str):
    """
    Modifies Indy request by adding information how to pay fees for this transaction
    according to this payment method.

    This method consumes set of inputs and outputs. The difference between inputs balance
    and outputs balance is the fee for this transaction.

    Not that this method also produces correct fee signatures.

    Format of inputs is specific for payment method. Usually it should reference payment transaction
    with at least one output that corresponds to payment address that user owns.

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param req_json : initial transaction request as json
    :param inputs_json: The list of payment sources as json array:
      ["source1", ...]
        - each input should reference paymentAddress
        - this param will be used to determine payment_method
    :param outputs_json: The list of outputs as json array:
      [{
        recipient: <str>, // payment address of recipient
        amount: <int>, // amount
      }]
    :param extra: // optional information for payment operation

    :return:
        req_with_fees_json: modified Indy request with added fees info,
        payment_method: used payment method
    """

    logger = logging.getLogger(__name__)
    logger.debug(
        "add_request_fees: >>> wallet_handle: %r, submitter_did: %r, req_json: %r, inputs_json: %r, outputs_json: %r, "
        "extra: %r",
        wallet_handle,
        submitter_did,
        req_json,
        inputs_json,
        outputs_json,
        extra)

    if not hasattr(add_request_fees, "cb"):
        logger.debug("add_request_fees: Creating callback")
        add_request_fees.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_req_json = c_char_p(req_json.encode('utf-8'))
    c_inputs_json = c_char_p(inputs_json.encode('utf-8'))
    c_outputs_json = c_char_p(outputs_json.encode('utf-8'))
    c_extra = c_char_p(extra.encode('utf-8')) if extra is not None else None

    (req_with_fees_json, payment_method) = await do_call('indy_add_request_fees',
                                                         c_wallet_handle,
                                                         c_submitter_did,
                                                         c_req_json,
                                                         c_inputs_json,
                                                         c_outputs_json,
                                                         c_extra,
                                                         add_request_fees.cb)
    res = (req_with_fees_json.decode(), payment_method.decode())

    logger.debug("add_request_fees: <<< res: %r", res)
    return res


async def parse_response_with_fees(payment_method: str,
                                   resp_json: str) -> str:
    """
    Parses response for Indy request with fees.

    :param payment_method: Payment method to use (for example, 'sov').
    :param resp_json: response for Indy request with fees
               Note: this param will be used to determine payment_method
    :return: receipts_json - parsed (payment method and node version agnostic) receipts info as json:
      [{
         receipt: <str>, // receipt that can be used for payment referencing and verification
         recipient: <str>, //payment address of recipient
         amount: <int>, // amount
         extra: <str>, // optional data from payment transaction
      }]
    """

    logger = logging.getLogger(__name__)
    logger.debug("parse_response_with_fees: >>> payment_method: %r, resp_json: %r",
                 payment_method,
                 resp_json)

    if not hasattr(parse_response_with_fees, "cb"):
        logger.debug("parse_response_with_fees: Creating callback")
        parse_response_with_fees.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_resp_json = c_char_p(resp_json.encode('utf-8'))

    receipts_json = await do_call('indy_parse_response_with_fees',
                                  c_payment_method,
                                  c_resp_json,
                                  parse_response_with_fees.cb)

    res = receipts_json.decode()
    logger.debug("parse_response_with_fees: <<< res: %r", res)
    return res


async def build_get_payment_sources_request(wallet_handle: int,
                                            submitter_did: str,
                                            payment_address: str) -> (str, str):
    """
    Builds Indy request for getting sources list for payment address
    according to this payment method.
    Deprecated. This function will be most likely be removed with Indy SDK 2.0 version

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param payment_address: target payment address
    :return: get_sources_txn_json: Indy request for getting sources list for payment address
             payment_method: used payment method
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_payment_sources_request: >>> wallet_handle: %r, submitter_did: %r, payment_address: %r",
                 wallet_handle,
                 submitter_did,
                 payment_address)

    if not hasattr(build_get_payment_sources_request, "cb"):
        logger.debug("build_get_payment_sources_request: Creating callback")
        build_get_payment_sources_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_payment_address = c_char_p(payment_address.encode('utf-8'))

    (get_sources_txn_json, payment_method) = await do_call('indy_build_get_payment_sources_request',
                                                           c_wallet_handle,
                                                           c_submitter_did,
                                                           c_payment_address,
                                                           build_get_payment_sources_request.cb)
    res = (get_sources_txn_json.decode(), payment_method.decode())

    logger.debug("build_get_payment_sources_request: <<< res: %r", res)
    return res


async def build_get_payment_sources_with_from_request(wallet_handle: int,
                                                      submitter_did: str,
                                                      payment_address: str,
                                                      from_seqno: int = -1) -> (str, str):
    """
    Builds Indy request for getting sources list for payment address
    according to this payment method.

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param payment_address: target payment address
    :param from_seqno: shift to the next slice of payment sources
    :return: get_sources_txn_json: Indy request for getting sources list for payment address
             payment_method: used payment method
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_payment_sources_with_from_request: >>> wallet_handle: %r, submitter_did: %r, payment_address: %r",
                 wallet_handle,
                 submitter_did,
                 payment_address)

    if not hasattr(build_get_payment_sources_with_from_request, "cb"):
        logger.debug("build_get_payment_sources_with_from_request: Creating callback")
        build_get_payment_sources_with_from_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_payment_address = c_char_p(payment_address.encode('utf-8'))
    c_from_seqno = c_int64(from_seqno)

    (get_sources_txn_json, payment_method) = await do_call('indy_build_get_payment_sources_with_from_request',
                                                           c_wallet_handle,
                                                           c_submitter_did,
                                                           c_payment_address,
                                                           c_from_seqno,
                                                           build_get_payment_sources_with_from_request.cb)
    res = (get_sources_txn_json.decode(), payment_method.decode())

    logger.debug("build_get_payment_sources_with_from_request: <<< res: %r", res)
    return res


async def parse_get_payment_sources_with_from_response(payment_method: str,
                                                       resp_json: str) -> str:
    """
    Parses response for Indy request for getting sources list.

    :param payment_method: Payment method to use (for example, 'sov').
    :param resp_json: resp_json: response for Indy request for getting sources list
                      Note: this param will be used to determine payment_method
    :return: sources_json: parsed (payment method and node version agnostic) sources info as json:
      [{
         source: <str>, // source input
         paymentAddress: <str>, //payment address for this source
         amount: <int>, // amount
         extra: <str>, // optional data from payment transaction
      }],
      next: pointer to the next slice of payment sources
    """

    logger = logging.getLogger(__name__)
    logger.debug("parse_get_payment_sources_with_fromresponse: >>> payment_method: %r, resp_json: %r",
                 payment_method,
                 resp_json)

    if not hasattr(parse_get_payment_sources_with_from_response, "cb"):
        logger.debug("parse_get_payment_sources_with_from_response: Creating callback")
        parse_get_payment_sources_with_from_response.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_int64))

    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_resp_json = c_char_p(resp_json.encode('utf-8'))

    sources_json, next_seqno = await do_call('indy_parse_get_payment_sources_with_from_response',
                                       c_payment_method,
                                       c_resp_json,
                                       parse_get_payment_sources_with_from_response.cb)

    res = sources_json.decode()
    logger.debug("parse_get_payment_sources_with_from_response: <<< res: %r", res)
    return res, next_seqno


async def parse_get_payment_sources_response(payment_method: str,
                                             resp_json: str) -> str:
    """
    Parses response for Indy request for getting sources list.

    :param payment_method: Payment method to use (for example, 'sov').
    :param resp_json: resp_json: response for Indy request for getting sources list
                      Note: this param will be used to determine payment_method
    :return: sources_json: parsed (payment method and node version agnostic) sources info as json:
      [{
         source: <str>, // source input
         paymentAddress: <str>, //payment address for this source
         amount: <int>, // amount
         extra: <str>, // optional data from payment transaction
      }]
    """

    logger = logging.getLogger(__name__)
    logger.debug("parse_get_payment_sources_response: >>> payment_method: %r, resp_json: %r",
                 payment_method,
                 resp_json)

    if not hasattr(parse_get_payment_sources_response, "cb"):
        logger.debug("parse_get_payment_sources_response: Creating callback")
        parse_get_payment_sources_response.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_resp_json = c_char_p(resp_json.encode('utf-8'))

    sources_json = await do_call('indy_parse_get_payment_sources_response',
                                 c_payment_method,
                                 c_resp_json,
                                 parse_get_payment_sources_response.cb)

    res = sources_json.decode()
    logger.debug("parse_get_payment_sources_response: <<< res: %r", res)
    return res


async def build_payment_req(wallet_handle: int,
                            submitter_did: str,
                            inputs_json: str,
                            outputs_json: str,
                            extra: Optional[str]) -> (str, str):
    """
    Builds Indy request for doing payment
    according to this payment method.
   
    This method consumes set of inputs and outputs.
   
    Format of inputs is specific for payment method. Usually it should reference payment transaction
    with at least one output that corresponds to payment address that user owns.

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param inputs_json: The list of payment sources as json array:
      ["source1", ...]
      Note that each source should reference payment address    
    :param outputs_json: The list of outputs as json array:
      [{
        recipient: <str>, // payment address of recipient
        amount: <int>, // amount
      }]
    :param extra: // optional information for payment operation

    :return: payment_req_json: Indy request for doing payment
             payment_method: used payment method
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_payment_req: >>> wallet_handle: %r, submitter_did: %r, inputs_json: %r, outputs_json: %r,"
                 " extra: %r",
                 wallet_handle,
                 submitter_did,
                 inputs_json,
                 outputs_json,
                 extra)

    if not hasattr(build_payment_req, "cb"):
        logger.debug("build_payment_req: Creating callback")
        build_payment_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_inputs_json = c_char_p(inputs_json.encode('utf-8'))
    c_outputs_json = c_char_p(outputs_json.encode('utf-8'))
    c_extra = c_char_p(extra.encode('utf-8')) if extra is not None else None

    (payment_req_json, payment_method) = await do_call('indy_build_payment_req',
                                                       c_wallet_handle,
                                                       c_submitter_did,
                                                       c_inputs_json,
                                                       c_outputs_json,
                                                       c_extra,
                                                       build_payment_req.cb)
    res = (payment_req_json.decode(), payment_method.decode())

    logger.debug("build_payment_req: <<< res: %r", res)
    return res


async def parse_payment_response(payment_method: str,
                                 resp_json: str) -> str:
    """
    Parses response for Indy request for payment txn.

    :param payment_method: Payment method to use (for example, 'sov').
    :param resp_json: resp_json: response for Indy request for payment txn
      Note: this param will be used to determine payment_method
    :return: receipts_json: parsed (payment method and node version agnostic) receipts info as json:
      [{
         receipt: <str>, // receipt that can be used for payment referencing and verification
         recipient: <str>, // payment address of recipient
         amount: <int>, // amount
         extra: <str>, // optional data from payment transaction
      }]
    """

    logger = logging.getLogger(__name__)
    logger.debug("parse_payment_response: >>> wallet_handle: %r, payment_method: %r, resp_json: %r",
                 payment_method,
                 resp_json)

    if not hasattr(parse_payment_response, "cb"):
        logger.debug("parse_payment_response: Creating callback")
        parse_payment_response.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_resp_json = c_char_p(resp_json.encode('utf-8'))

    receipts_json = await do_call('indy_parse_payment_response',
                                  c_payment_method,
                                  c_resp_json,
                                  parse_payment_response.cb)

    res = receipts_json.decode()
    logger.debug("parse_payment_response: <<< res: %r", res)
    return res


async def prepare_payment_extra_with_acceptance_data(extra_json: Optional[str],
                                                     text: Optional[str],
                                                     version: Optional[str],
                                                     taa_digest: Optional[str],
                                                     mechanism: str,
                                                     time: int) -> str:
    """
    Append payment extra JSON with TAA acceptance data
   
    EXPERIMENTAL
   
    This function may calculate digest by itself or consume it as a parameter.
    If all text, version and taa_digest parameters are specified, a check integrity of them will be done.

    :param extra_json: (Optional) original extra json.
    :param text and version: (Optional) raw data about TAA from ledger.
               These parameters should be passed together.
               These parameters are required if taa_digest parameter is omitted.
    :param taa_digest: (Optional) digest on text and version.
                       Digest is sha256 hash calculated on concatenated strings: version || text.
                       This parameter is required if text and version parameters are omitted.
    :param mechanism: mechanism how user has accepted the TAA
    :param time: UTC timestamp when user has accepted the TAA

    :return: Updated request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug(
        "prepare_payment_extra_with_acceptance_data: >>> extra_json: %r, text: %r, version: %r, hash: %r, "
        "acc_mech_type: %r, time_of_acceptance: %r",
        extra_json,
        text,
        version,
        taa_digest,
        mechanism,
        time)

    if not hasattr(prepare_payment_extra_with_acceptance_data, "cb"):
        logger.debug("prepare_payment_extra_with_acceptance_data: Creating callback")
        prepare_payment_extra_with_acceptance_data.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_extra_json = c_char_p(extra_json.encode('utf-8')) if extra_json is not None else None
    c_text = c_char_p(text.encode('utf-8')) if text is not None else None
    c_version = c_char_p(version.encode('utf-8')) if version is not None else None
    c_taa_digest = c_char_p(taa_digest.encode('utf-8')) if taa_digest is not None else None
    c_mechanism = c_char_p(mechanism.encode('utf-8'))

    request_json = await do_call('indy_prepare_payment_extra_with_acceptance_data',
                                 c_extra_json,
                                 c_text,
                                 c_version,
                                 c_taa_digest,
                                 c_mechanism,
                                 c_uint64(time),
                                 prepare_payment_extra_with_acceptance_data.cb)

    res = request_json.decode()
    logger.debug("prepare_payment_extra_with_acceptance_data: <<< res: %r", res)
    return res


async def build_mint_req(wallet_handle: int,
                         submitter_did: str,
                         outputs_json: str,
                         extra: Optional[str]) -> (str, str):
    """
    Builds Indy request for doing minting
    according to this payment method.

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param outputs_json: The list of outputs as json array:
      [{
        recipient: <str>, // payment address of recipient
        amount: <int>, // amount
        extra: <str>, // optional data
      }]
    :param extra: // optional information for payment operation

    :return: mint_req_json: Indy request for doing minting
             payment_method: used payment method
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_mint_req: >>> wallet_handle: %r, submitter_did: %r, outputs_json: %r, extra: %r",
                 wallet_handle,
                 submitter_did,
                 outputs_json,
                 extra)

    if not hasattr(build_mint_req, "cb"):
        logger.debug("build_mint_req: Creating callback")
        build_mint_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_outputs_json = c_char_p(outputs_json.encode('utf-8'))
    c_extra = c_char_p(extra.encode('utf-8')) if extra is not None else None

    (mint_req_json, payment_method) = await do_call('indy_build_mint_req',
                                                    c_wallet_handle,
                                                    c_submitter_did,
                                                    c_outputs_json,
                                                    c_extra,
                                                    build_mint_req.cb)
    res = (mint_req_json.decode(), payment_method.decode())

    logger.debug("build_mint_req: <<< res: %r", res)
    return res


async def build_set_txn_fees_req(wallet_handle: int,
                                 submitter_did: str,
                                 payment_method: str,
                                 fees_json: str) -> str:
    """
    Builds Indy request for setting fees for transactions in the ledger

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param payment_method: Payment method to use (for example, 'sov').
    :param fees_json: {
       txnType1: amount1,
       txnType2: amount2,
       .................
       txnTypeN: amountN,
     }
    :return: set_txn_fees_json: Indy request for setting fees for transactions in the ledger
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_set_txn_fees_req: >>> wallet_handle: %r, submitter_did: %r, payment_method: %r, fees_json: %r",
                 wallet_handle,
                 submitter_did,
                 payment_method,
                 fees_json)

    if not hasattr(build_set_txn_fees_req, "cb"):
        logger.debug("build_set_txn_fees_req: Creating callback")
        build_set_txn_fees_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_fees_json = c_char_p(fees_json.encode('utf-8'))

    set_txn_fees_json = await do_call('indy_build_set_txn_fees_req',
                                      c_wallet_handle,
                                      c_submitter_did,
                                      c_payment_method,
                                      c_fees_json,
                                      build_set_txn_fees_req.cb)

    res = set_txn_fees_json.decode()
    logger.debug("build_set_txn_fees_req: <<< res: %r", res)
    return res


async def build_get_txn_fees_req(wallet_handle: int,
                                 submitter_did: str,
                                 payment_method: str) -> str:
    """
    Builds Indy request for getting fees for transactions in the ledger

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param payment_method: Payment method to use (for example, 'sov').
    :return: set_txn_fees_json: Indy request for setting fees for transactions in the ledger
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_txn_fees_req: >>> wallet_handle: %r, submitter_did: %r, payment_method: %r",
                 wallet_handle,
                 submitter_did,
                 payment_method)

    if not hasattr(build_get_txn_fees_req, "cb"):
        logger.debug("build_get_txn_fees_req: Creating callback")
        build_get_txn_fees_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_payment_method = c_char_p(payment_method.encode('utf-8'))

    get_txn_fees_json = await do_call('indy_build_get_txn_fees_req',
                                      c_wallet_handle,
                                      c_submitter_did,
                                      c_payment_method,
                                      build_get_txn_fees_req.cb)

    res = get_txn_fees_json.decode()
    logger.debug("build_get_txn_fees_req: <<< res: %r", res)
    return res


async def parse_get_txn_fees_response(payment_method: str,
                                      resp_json: str) -> str:
    """
    Parses response for Indy request for getting fees

    :param payment_method: Payment method to use (for example, 'sov').
    :param resp_json: response for Indy request for getting fees
    :return: fees_json: {
       txnType1: amount1,
       txnType2: amount2,
       .................
       txnTypeN: amountN,
     }
    """

    logger = logging.getLogger(__name__)
    logger.debug("parse_get_txn_fees_response: >>> payment_method: %r, resp_json: %r",
                 payment_method,
                 resp_json)

    if not hasattr(parse_get_txn_fees_response, "cb"):
        logger.debug("parse_get_txn_fees_response: Creating callback")
        parse_get_txn_fees_response.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_resp_json = c_char_p(resp_json.encode('utf-8'))

    fees_json = await do_call('indy_parse_get_txn_fees_response',
                              c_payment_method,
                              c_resp_json,
                              parse_get_txn_fees_response.cb)

    res = fees_json.decode()
    logger.debug("parse_get_txn_fees_response: <<< res: %r", res)
    return res


async def build_verify_payment_req(wallet_handle: int,
                                   submitter_did: str,
                                   receipt: str) -> (str, str):
    """
    Builds Indy request for information to verify the payment receipt

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did : (Option) DID of request sender
    :param receipt: payment receipt to verify

    :return: verify_txn_json: Indy request for verification receipt for transactions in the ledger
             payment_method: used payment method
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_verify_payment_req: >>> wallet_handle: %r, submitter_did: %r, receipt: %r",
                 wallet_handle,
                 submitter_did,
                 receipt)

    if not hasattr(build_verify_payment_req, "cb"):
        logger.debug("build_verify_payment_req: Creating callback")
        build_verify_payment_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8')) if submitter_did is not None else None
    c_receipt = c_char_p(receipt.encode('utf-8'))

    (verify_txn_json, payment_method) = await do_call('indy_build_verify_payment_req',
                                                      c_wallet_handle,
                                                      c_submitter_did,
                                                      c_receipt,
                                                      build_verify_payment_req.cb)
    res = (verify_txn_json.decode(), payment_method.decode())

    logger.debug("build_verify_payment_req: <<< res: %r", res)
    return res


async def parse_verify_payment_response(payment_method: str,
                                        resp_json: str) -> str:
    """
    Parses Indy response with information to verify receipt

    :param payment_method: Payment method to use (for example, 'sov').
    :param resp_json: resp_json: response for Indy request for payment txn
      Note: this param will be used to determine payment_method
    :return: receipts_json: parsed (payment method and node version agnostic) receipt verification info as json:
    {
         sources: [<str>, ]
         receipts: [ {
             recipient: <str>, // payment address of recipient
             receipt: <str>, // receipt that can be used for payment referencing and verification
             amount: <int>, // amount
         } ],
         extra: <str>, //optional data
    }
    """

    logger = logging.getLogger(__name__)
    logger.debug("parse_verify_payment_response: >>> wallet_handle: %r, payment_method: %r, resp_json: %r",
                 payment_method,
                 resp_json)

    if not hasattr(parse_verify_payment_response, "cb"):
        logger.debug("parse_verify_payment_response: Creating callback")
        parse_verify_payment_response.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_payment_method = c_char_p(payment_method.encode('utf-8'))
    c_resp_json = c_char_p(resp_json.encode('utf-8'))

    receipts_json = await do_call('indy_parse_verify_payment_response',
                                  c_payment_method,
                                  c_resp_json,
                                  parse_verify_payment_response.cb)

    res = receipts_json.decode()
    logger.debug("parse_verify_payment_response: <<< res: %r", res)
    return res


async def get_request_info(get_auth_rule_response_json: str,
                           requester_info_json: str,
                           fees_json: str) -> str:
    """
    Gets request requirements (with minimal price) correspondent to specific auth rule
    in case the requester can perform this action.
 
    EXPERIMENTAL
 
    If the requester does not match to the request constraints `TransactionNotAllowed` error will be thrown.

    :param get_auth_rule_response_json: response on GET_AUTH_RULE request returning action constraints set on the ledger.
    :param requester_info_json: {
        "role": string (optional) - role of a user which can sign a transaction.
        "sig_count": u64 - number of signers.
        "is_owner": bool (optional) - if user is an owner of transaction (false by default).
        "is_off_ledger_signature": bool (optional) - if user did is unknow for ledger (false by default).
    }
    :param fees_json: fees set on the ledger (result of `parse_get_txn_fees_response`).

    :return: request_info_json: request info if a requester match to the action constraints.
    {
       "price": u64 - fee required for the action performing,
       "requirements": [{
           "role": string (optional) - role of users who should sign,
           "sig_count": u64 - number of signers,
           "need_to_be_owner": bool (optional) - if requester need to be owner,
           "off_ledger_signature": bool (optional) - allow signature of unknow for ledger did (false by default).
       }]
    }
    """

    logger = logging.getLogger(__name__)
    logger.debug("get_request_info: >>> get_auth_rule_response_json: %r, requester_info_json: %r, fees_json: %r",
                 get_auth_rule_response_json,
                 requester_info_json,
                 fees_json)

    if not hasattr(get_request_info, "cb"):
        logger.debug("get_request_info: Creating callback")
        get_request_info.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_get_auth_rule_response_json = c_char_p(get_auth_rule_response_json.encode('utf-8'))
    c_requester_info_json = c_char_p(requester_info_json.encode('utf-8'))
    c_fees_json = c_char_p(fees_json.encode('utf-8'))

    request_info_json = await do_call('indy_get_request_info',
                                      c_get_auth_rule_response_json,
                                      c_requester_info_json,
                                      c_fees_json,
                                      get_request_info.cb)

    res = request_info_json.decode()
    logger.debug("get_request_info: <<< res: %r", res)
    return res


async def sign_with_address(wallet_handle: int,
                            address: str,
                            msg: bytes) -> bytes:
    """
    Signs a message with a payment address.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param address:  payment address of message signer. The key must be created by calling create_payment_address
    :param msg: a message to be signed
    :return: a signature string
    """

    logger = logging.getLogger(__name__)
    logger.debug("sign_with_address: >>> wallet_handle: %r, address: %r, msg: %r",
                 wallet_handle,
                 address,
                 msg)


    def transform_cb(arr_ptr: POINTER(c_uint8), arr_len: c_uint32):
        return bytes(arr_ptr[:arr_len]),


    if not hasattr(sign_with_address, "cb"):
        logger.debug("sign_with_address: Creating callback")
        sign_with_address.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, POINTER(c_uint8), c_uint32), transform_cb)

    c_wallet_handle = c_int32(wallet_handle)
    c_address = c_char_p(address.encode('utf-8'))
    c_msg_len = c_uint32(len(msg))

    signature = await do_call('indy_sign_with_address',
                              c_wallet_handle,
                              c_address,
                              msg,
                              c_msg_len,
                              sign_with_address.cb)

    logger.debug("sign_with_address: <<< res: %r", signature)
    return signature


async def verify_with_address(address: str,
                              msg: bytes,
                              signature: bytes) -> bool:
    """
    Verify a signature with a payment address.

    :param address: payment address of the message signer
    :param msg: message that has been signed
    :param signature: a signature to be verified
    :return: valid: true - if signature is valid, false - otherwise
    """

    logger = logging.getLogger(__name__)
    logger.debug("verify_with_address: >>> address: %r, signed_msg: %r, signature: %r",
                 address,
                 msg,
                 signature)

    if not hasattr(verify_with_address, "cb"):
        logger.debug("verify_with_address: Creating callback")
        verify_with_address.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_bool))

    c_address = c_char_p(address.encode('utf-8'))
    c_msg_len = c_uint32(len(msg))
    c_signature_len = c_uint32(len(signature))

    res = await do_call('indy_verify_with_address',
                        c_address,
                        msg,
                        c_msg_len,
                        signature,
                        c_signature_len,
                        verify_with_address.cb)

    logger.debug("verify_with_address: <<< res: %r", res)
    return res



from .libindy import do_call, create_cb

from ctypes import *
from typing import Optional

import logging


async def is_pairwise_exists(wallet_handle: int,
                             their_did: str) -> bool:
    """
    Check if pairwise is exists.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param their_did: encoded Did.
    :return: true - if pairwise is exists, false - otherwise
    """

    logger = logging.getLogger(__name__)
    logger.debug("is_pairwise_exists: >>> wallet_handle: %r, their_did: %r",
                 wallet_handle,
                 their_did)

    if not hasattr(is_pairwise_exists, "cb"):
        logger.debug("is_pairwise_exists: Creating callback")
        is_pairwise_exists.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_bool))

    c_wallet_handle = c_int32(wallet_handle)
    c_their_did = c_char_p(their_did.encode('utf-8'))

    res = await do_call('indy_is_pairwise_exists',
                        c_wallet_handle,
                        c_their_did,
                        is_pairwise_exists.cb)

    logger.debug("is_pairwise_exists: <<< res: %r", res)
    return res


async def create_pairwise(wallet_handle: int,
                          their_did: str,
                          my_did: str,
                          metadata: Optional[str]) -> None:
    """
    Creates pairwise.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param their_did: encrypting DID
    :param my_did: encrypting DID
    :param metadata: (Optional) extra information for pairwise
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_pairwise: >>> wallet_handle: %r, their_did: %r, my_did: %r, metadata: %r",
                 wallet_handle,
                 their_did,
                 my_did,
                 metadata)

    if not hasattr(create_pairwise, "cb"):
        logger.debug("create_pairwise: Creating callback")
        create_pairwise.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_their_did = c_char_p(their_did.encode('utf-8'))
    c_my_did = c_char_p(my_did.encode('utf-8'))
    c_metadata = c_char_p(metadata.encode('utf-8')) if metadata is not None else None

    await do_call('indy_create_pairwise',
                  c_wallet_handle,
                  c_their_did,
                  c_my_did,
                  c_metadata,
                  create_pairwise.cb)

    logger.debug("create_pairwise: <<<")


async def list_pairwise(wallet_handle: int) -> str:
    """
    Get list of saved pairwise.

    :param wallet_handle: wallet handler (created by open_wallet).
    :return: pairwise_list: list of saved pairwise
    """

    logger = logging.getLogger(__name__)
    logger.debug("list_pairwise: >>> wallet_handle: %r", wallet_handle)

    if not hasattr(list_pairwise, "cb"):
        logger.debug("list_pairwise: Creating callback")
        list_pairwise.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)

    pairwise_list = await do_call('indy_list_pairwise',
                                  c_wallet_handle,
                                  list_pairwise.cb)

    res = pairwise_list.decode()
    logger.debug("list_pairwise: <<< res: %r", res)
    return res


async def get_pairwise(wallet_handle: int,
                       their_did: str) -> None:
    """
    Gets pairwise information for specific their_did.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param their_did: encoded Did
    :return: pairwise_info_json: did info associated with their did
    """

    logger = logging.getLogger(__name__)
    logger.debug("get_pairwise: >>> wallet_handle: %r, their_did: %r",
                 wallet_handle,
                 their_did)

    if not hasattr(get_pairwise, "cb"):
        logger.debug("get_pairwise: Creating callback")
        get_pairwise.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_their_did = c_char_p(their_did.encode('utf-8'))

    pairwise_info_json = await do_call('indy_get_pairwise',
                                       c_wallet_handle,
                                       c_their_did,
                                       get_pairwise.cb)

    res = pairwise_info_json.decode()
    logger.debug("get_pairwise: <<< res: %r", res)
    return res


async def set_pairwise_metadata(wallet_handle: int,
                                their_did: str,
                                metadata: Optional[str]) -> None:
    """
    Save some data in the Wallet for pairwise associated with Did.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param their_did: encoded DID
    :param metadata: some extra information for pairwise
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("set_pairwise_metadata: >>> wallet_handle: %r, their_did: %r, metadata: %r",
                 wallet_handle,
                 their_did,
                 metadata)

    if not hasattr(set_pairwise_metadata, "cb"):
        logger.debug("set_pairwise_metadata: Creating callback")
        set_pairwise_metadata.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_their_did = c_char_p(their_did.encode('utf-8'))
    c_metadata = c_char_p(metadata.encode('utf-8')) if metadata is not None else None

    await do_call('indy_set_pairwise_metadata',
                  c_wallet_handle,
                  c_their_did,
                  c_metadata,
                  set_pairwise_metadata.cb)

    logger.debug("set_pairwise_metadata: <<<")
